#include "apu.h"
#include "main.h"

void Capu::ClockTriangleSequencer()
{
u8 t_step[32] =
	{0xF,0xE,0xD,0xC,0xB,0xA,9,8,7,6,5,4,3,2,1,0,0,1,2,3,4,5,6,7,8,9,0xA,0xB,0xC,0xD,0xE,0xF};

t_sequencerpos++;
t_sequencerpos &= 0x1F;
t_output = t_step[t_sequencerpos];
}

void Capu::ClockTriangleLinearCounter()
{
if(t_haltlinear)						//if halt flag is set reload counter
	t_linearcounter = regs[8] & 0x7F;
else
	{
	if(t_linearcounter)
		t_linearcounter--;
	else
		{}
	}
if((regs[8] & 0x80) == 0)			//if control flag is clear, clear halt flag
	t_haltlinear = 0;
}

void Capu::ClockTriangleLengthCounter()
{
if((regs[8] & 0x80) == 0)
	{
	if(t_lengthcounter)
		t_lengthcounter--;
	}
}

void Capu::ClockTriangleTimer()
{
if(t_timer == 0)
	{
	t_timer = regs[0xA] | ((regs[0xB] & 7) << 8);
	if(t_linearcounter && t_lengthcounter)
		ClockTriangleSequencer();
	}
t_timer--;
}

void Capu::ClockFrameCounter()
{
if(regs[0x17] & 0x80)				//mode 1, 5 step
	{
	switch(f_sequencerpos)
		{
		case 0:							//"clock length counters and sweep units"
		case 2:
			ClockTriangleLengthCounter();
		case 1:							//"clock envelopes and triangle's linear counter"
		case 3:
			ClockTriangleLinearCounter();
			break;
		case 4:
			break;
		}
	f_sequencerpos++;
	if(f_sequencerpos == 5)
		f_sequencerpos = 0;
	}
else										//mode 0, 4 step
	{
	switch(f_sequencerpos)
		{
		case 1:							//"clock length counters and sweep units"
		case 3:
			ClockTriangleLengthCounter();
		case 0:							//"clock envelopes and triangle's linear counter"
		case 2:
			ClockTriangleLinearCounter();
			break;
		}
	f_sequencerpos++;
	f_sequencerpos &= 3;
	}

}

void Capu::FillSamples(int n)
{
int nn[5] = {0,184,368,551,735};
u8 *buf;
int i;
u8 output,square_out,tnd_out;

//square_out = (u8)(95.88f/(float)((8128/(s1_output+s2_output))+100));
//tnd_out = (u8)(159.79f/(float)((1/((t_output/8227)+(n_output/12241)+(d_output/22638)))+100));
square_out = square_table[s1_output + s2_output];
tnd_out = tnd_table[3 * t_output + 2 * n_output + d_output];
output = square_out + tnd_out;
buf = soundbuf + nn[n];
for(i=nn[n];i<nn[n+1];i++)
	*buf++ = output;
}

Capu::Capu()							//initialize tables
{
int i;

for(i=0;i<32;i++)
	square_table[i] = (u8)(95.52f / (8128.0f / (float)i + 100.0f));
for(i=0;i<256;i++)
	tnd_table[i] = (u8)(163.67f / (24329.0f / (float)i + 100.0f));
}

void Capu::Reset()					//reset 2a03 sound hardware
{
t_step = 0;
t_output = s1_output = s2_output = n_output = d_output = 0;
}

void Capu::Step()						//execute one apu cycle (1/3 the speed of ppu) 1.79mhz
{
if(cycle == 7457)
	{
	ClockFrameCounter();
	FillSamples(0);
	}
else if(cycle == 14915)
	{
	ClockFrameCounter();
	FillSamples(1);
	}
else if(cycle == 22372)
	{
	ClockFrameCounter();
	FillSamples(2);
	}
else if(cycle == 29829)
	{
	ClockFrameCounter();
	FillSamples(3);
	cycle = 0;
	}
else
	cycle++;
ClockTriangleTimer();				//clocked on every cpu cycle
}

u8 Capu::Read(u32 addr)
{
return(0);
}

void Capu::Write(u32 addr,u8 data)
{
switch(addr)
	{
	case 0x4008:
		//message("$%04X write = $%02X\n",addr,data);
		regs[0x8] = data;
		t_linearcounter = data & 0x7F;
		break;
	case 0x400A:
		//message("$%04X write = $%02X\n",addr,data);
		regs[0xA] = data;
		t_timer &= 0x700;
		t_timer |= data;
		break;
	case 0x400B:
		//message("$%04X write = $%02X\n",addr,data);
		regs[0xB] = data;
		t_haltlinear = 1;
		t_timer &= 0xFF;
		t_timer |= (data & 7) << 8;
		if(data & 8)
			{
			if((data & 0xF0) == 0)
				t_lengthcounter = 0x7F;
			else
				t_lengthcounter = (data & 0xF0) >> 4;
			}
		else
			{
			switch(data & 0xF0)
				{
				case 0x00:t_lengthcounter = 0x05;break;
				case 0x10:t_lengthcounter = 0x0A;break;
				case 0x20:t_lengthcounter = 0x14;break;
				case 0x30:t_lengthcounter = 0x28;break;
				case 0x40:t_lengthcounter = 0x50;break;
				case 0x50:t_lengthcounter = 0x1E;break;
				case 0x60:t_lengthcounter = 0x07;break;
				case 0x70:t_lengthcounter = 0x0D;break;
				case 0x80:t_lengthcounter = 0x06;break;
				case 0x90:t_lengthcounter = 0x0C;break;
				case 0xA0:t_lengthcounter = 0x18;break;
				case 0xB0:t_lengthcounter = 0x30;break;
				case 0xC0:t_lengthcounter = 0x60;break;
				case 0xD0:t_lengthcounter = 0x24;break;
				case 0xE0:t_lengthcounter = 0x08;break;
				case 0xF0:t_lengthcounter = 0x10;break;
				}
			}
		break;
	case 0x4017:
		//message("$%04X write = $%02X\n",addr,data);
		regs[0x17] = data;
		if(data & 0x80)
			{
			f_sequencerpos = 0;
			ClockFrameCounter();
			}
		else
			f_sequencerpos = 4;
		break;
	}
}
