#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include "debuggerdlg.h"
#include "resource.h"
#include "main.h"

#define LINES	25 //lines in the list box

typedef struct breakpoint_s
	{
	u32 addr;
	u32 opcode;
	u32 type;
	} breakpoint_t;
	
HWND debuggerdlgwnd = 0;
breakpoint_t breakpoints[128];
int bpnum;

static void bp_add(u32 addr,u32 op,u32 type)
{
breakpoints[bpnum].addr = addr;
breakpoints[bpnum].opcode = op;
breakpoints[bpnum++].type = type;
}

static void bp_del(int idx)
{
int n;
breakpoint_t *bp = &breakpoints[idx],*nextbp;

bpnum--;
for(n=idx;n<bpnum;n++,bp++)
	{
	nextbp = bp + 1;
	bp->addr = nextbp->addr;
	bp->opcode = nextbp->opcode;
	bp->type = nextbp->type;
	}
}

static char *getbase(char *str,int *base)
{
*base = 10;
if(*str == '$')
	{
	str++;
	*base = 16;
	}
else if(*str == '0' && ((*(str+1) == 'x') || (*(str+1) == 'X')))
	{
	str += 2;
	*base = 16;
	}
else
	{
	char *s = str;
	
	while(*s)
		{
		if(tolower(*s) == 'a' || tolower(*s) == 'b' || tolower(*s) == 'c' || 
			tolower(*s) == 'd' || tolower(*s) == 'e' || tolower(*s) == 'f')
			{
			*base = 16;
			return(str);
			}
		s++;
		}
	}
return(str);
}
static u32 pc;

static char *getflags(u8 flags,char *s)
{
#define FLAG(v,l)	flags&v?l:'-'
sprintf(s,"%c%c%c%c%c%c%c",FLAG(0x80,'S'),FLAG(0x40,'V'),FLAG(0x10,'B'),FLAG(0x08,'D'),FLAG(0x04,'I'),FLAG(0x02,'Z'),FLAG(0x01,'C'));
return(s);
}

static void refresh(HWND dlg)
{
char opcodedis[64];
char str[128];
int i,line,savep;//,index;
u32 p;
char stuff[1024];

sprintf(stuff,
//cpu
	"\n$%04X\n$%02X\n$%02X\n$%02X\n$%02X\n$%02X\n\n"
//vectors
	"$%04X\n$%04X\n$%04X\n\n\n"
//ppu
	"$%02X\n$%02X\n$%02X\n%ld\n%ld\n%ld\n\n",

	nes.cpu->pc,nes.cpu->a,nes.cpu->x,nes.cpu->y,nes.cpu->s,nes.cpu->f,

	nes.cpu->MemRead(0xFFFA) | (nes.cpu->MemRead(0xFFFB) << 8),
	nes.cpu->MemRead(0xFFFE) | (nes.cpu->MemRead(0xFFFF) << 8),
	nes.cpu->MemRead(0xFFFC) | (nes.cpu->MemRead(0xFFFD) << 8),

	nes.ppu->control1,nes.ppu->control2,nes.ppu->status,
	nes.ppu->scanline,nes.ppu->pixel,
	/*nes.ppu->frames * (262*341) + */nes.ppu->framecycles
	);
SetDlgItemText(dlg,IDC_VALUES,stuff);
SendDlgItemMessage(dlg,IDC_DISASM,LB_RESETCONTENT,0,0);
line = -1;
p = pc;
for(i=0;i<LINES;i++)
	{
	if(p == nes.cpu->pc)
		line = i;
	savep = p;
	p = nes.cpu->Disassemble(opcodedis,p);
	sprintf(str,"%04X:\t%s",savep,opcodedis);
	SendDlgItemMessage(dlg,IDC_DISASM,LB_ADDSTRING,0,(LPARAM)(LPCTSTR)str);
	}
if(line != -1)
	SendDlgItemMessage(dlg,IDC_DISASM,LB_SETCURSEL,line,0);
SendDlgItemMessage(dlg,IDC_BPLIST,LB_RESETCONTENT,0,0);
for(i=0;i<bpnum;i++)
	{
	breakpoint_t *bp = &breakpoints[i];
			
	switch(bp->type)
		{
		case 1:sprintf(stuff,"Read $%04X",bp->addr);break;
		case 2:sprintf(stuff,"Write $%04X",bp->addr);break;
		case 4:sprintf(stuff,"Opcode $%02X",bp->opcode);break;
		case 8:sprintf(stuff,"NMI");break;
		case 0x10:sprintf(stuff,"IRQ");break;
		}
	SendDlgItemMessage(dlg,IDC_BPLIST,LB_ADDSTRING,0,(LPARAM)(LPCTSTR)stuff);	
	}
}

static int CALLBACK debuggerproc(HWND dlgwnd,UINT imsg,WPARAM wparam,LPARAM lparam)
{
char num[512];
u32 n;
char *stop,*p;
int b;

switch(imsg)
	{
	case WM_INITDIALOG:
		SetDlgItemText(dlgwnd,IDC_NAMES,"cpu:     \npc:\na:\nx:\ny:\ns:\nf:\n\nnmi:\nirq:\nreset:\n\nppu:     \nctrl1:\nctrl2\nstatus:\nline:\npixel:\nfcycle:");
		sprintf(num,"$%04X",nes.cpu->MemRead(0xFFFC) | (nes.cpu->MemRead(0xFFFD) << 8));
		SetDlgItemText(dlgwnd,IDC_JUMPADDR,num);
		SetDlgItemText(dlgwnd,IDC_STEPNUM,"1");
		SetDlgItemText(dlgwnd,IDC_ADDREDIT,"$2002");
		SetDlgItemText(dlgwnd,IDC_BREAKADDR,num);
		SendDlgItemMessage(dlgwnd,IDC_READRADIO,BM_SETCHECK,BST_CHECKED,0);	
		pc = nes.cpu->pc;
		refresh(dlgwnd);
		return(1);
	case WM_COMMAND:
		switch(LOWORD(wparam))
			{
			case IDC_RADIONMI:
				if(HIWORD(wparam) == BN_CLICKED)
					{
					
					return(1);
					}
				break;
			case IDC_ADDBTN:
				GetDlgItemText(dlgwnd,IDC_ADDREDIT,num,512);
				b = 10;
				p = getbase(num,&b);
				n = (u32)strtol(p,&stop,b);
				if(SendDlgItemMessage(dlgwnd,IDC_READRADIO,BM_GETCHECK,0,0) == BST_CHECKED)
					{
					bp_add(n,0,1);
					sprintf(num,"Read $%04X",n);
					}
				else if(SendDlgItemMessage(dlgwnd,IDC_WRITERADIO,BM_GETCHECK,0,0) == BST_CHECKED)
					{
					bp_add(n,0,2);
					sprintf(num,"Write $%04X",n);
					}
				else if(SendDlgItemMessage(dlgwnd,IDC_OPCODERADIO,BM_GETCHECK,0,0) == BST_CHECKED)
					{
					bp_add(0,n,4);
					sprintf(num,"Opcode $%02X",n);
					}
				else if(SendDlgItemMessage(dlgwnd,IDC_RADIONMI,BM_GETCHECK,0,0) == BST_CHECKED)
					{
					bp_add(n,0,8);
					sprintf(num,"NMI");
					}
				else if(SendDlgItemMessage(dlgwnd,IDC_RADIOIRQ,BM_GETCHECK,0,0) == BST_CHECKED)
					{
					bp_add(n,0,0x10);
					sprintf(num,"IRQ");
					}
				else
					return(1);
				SendDlgItemMessage(dlgwnd,IDC_BPLIST,LB_ADDSTRING,0,(LPARAM)(LPCTSTR)num);			
				return(1);
			case IDC_DELETEBTN:
				n = SendDlgItemMessage(dlgwnd,IDC_BPLIST,LB_GETCURSEL,0,0);
				if(n != LB_ERR)
					{
					bp_del(n);
					pc = nes.cpu->pc;
					refresh(dlgwnd);
					}
				return(1);
			case IDC_JUMP:
				GetDlgItemText(dlgwnd,IDC_JUMPADDR,num,512);
				if(strcmp(num,"") != 0)
					{
					char *stop,*p = num;
					int base = 10;

					p = getbase(p,&base);
					pc = (u16)strtol(p,&stop,base);
					refresh(dlgwnd);
					}
				return(1);
			case IDC_STEP:
				GetDlgItemText(dlgwnd,IDC_STEPNUM,num,512);
				if(strcmp(num,"") != 0)
					{
					char *stop,*p = num;
					int base = 10;
					u32 realcycles,n;

					p = getbase(p,&base);
					n = strtol(p,&stop,base);
					realcycles = 0;
					while(realcycles < n)
						{
						nes.Step();
						while(nes.cpu->opcycle != 1)
							nes.Step();
						realcycles++;
						}
					pc = nes.cpu->pc;
					refresh(dlgwnd);
					}
				return(1);
			case IDC_BREAKBTN:
				if(bpnum == 0)
					break;
				for(n=0;;n++)
					{
					int i,breakhit;
					breakpoint_t *bp = breakpoints;
					
					if(GetAsyncKeyState('Z'))
						{
						message("user break\n");
						break;
						}
					if((n & 0xFF) == 0)
						{
						pc = nes.cpu->pc;
						refresh(dlgwnd);
						}
					nes.Step();
					for(breakhit=0,i=0;i<bpnum;i++)
						{
						switch(bp->type)
							{
							case 1:
								if(nes.cpu->lastread == bp->addr)
									{
									message("breakpoint hit, address read $%04X\n",bp->addr);
									breakhit++;
									}
								break;
							case 2:
								if(nes.cpu->lastwrite == bp->addr)
									{
									message("breakpoint hit, address write $%04X\n",bp->addr);
									breakhit++;
									}
								break;
							case 4:
								if(nes.cpu->opcode == bp->opcode)
									{
									message("breakpoint hit, opcode $%02X\n",bp->opcode);
									breakhit++;
									}
								break;
							case 8:
								if(nes.cpu->nmi_active)
									{
									message("breakpoint hit, nmi\n");
									breakhit++;
									}
								break;
							case 0x10:
								if(nes.cpu->irq_active)
									{
									message("breakpoint hit, irq\n");
									breakhit++;
									}
								break;
							}
						}
					if(breakhit)
						break;
					}
				return(1);
			case IDC_REFRESH:
				pc = nes.cpu->pc;
				refresh(dlgwnd);
				return(1);
			case IDOK:
			case IDCANCEL:
				EndDialog(dlgwnd,0);
				return(1);
			}
		break;
	}
return(0);
} 

void debuggerdlg(HINSTANCE hinstance,HWND hwnd)
{
//todo: make modal
if((debuggerdlgwnd = CreateDialog(hinstance,MAKEINTRESOURCE(IDD_DEBUGGER),hwnd,debuggerproc)) != 0)
	{}
UpdateWindow(debuggerdlgwnd);
ShowWindow(debuggerdlgwnd,SW_SHOW);
}
