#include <windows.h>
#include "dx7video.h"
#include "main.h"

static u32 lockflags = 0;

void CDX7Video::ShowError(int error,char *func)
{
#define ERRORCASE(v)case v:message("%s: "#v"\n",func);break;
message("%s...",func);
switch(error)
	{
	ERRORCASE(DDERR_INCOMPATIBLEPRIMARY);
	ERRORCASE(DDERR_INVALIDCAPS);
	ERRORCASE(DDERR_INVALIDOBJECT);
	ERRORCASE(DDERR_INVALIDPARAMS);
	ERRORCASE(DDERR_INVALIDPIXELFORMAT);
	ERRORCASE(DDERR_NOALPHAHW);
	ERRORCASE(DDERR_NOCOOPERATIVELEVELSET);
	ERRORCASE(DDERR_NODIRECTDRAWHW);
	ERRORCASE(DDERR_NOEMULATION);
	ERRORCASE(DDERR_NOEXCLUSIVEMODE);
	ERRORCASE(DDERR_NOFLIPHW);
	ERRORCASE(DDERR_NOMIPMAPHW);
	ERRORCASE(DDERR_NOOVERLAYHW);
	ERRORCASE(DDERR_NOZBUFFERHW);
	ERRORCASE(DDERR_OUTOFMEMORY);
	ERRORCASE(DDERR_OUTOFVIDEOMEMORY);
	ERRORCASE(DDERR_PRIMARYSURFACEALREADYEXISTS);
	ERRORCASE(DDERR_UNSUPPORTEDMODE);
	ERRORCASE(DDERR_UNSUPPORTED);
	ERRORCASE(DDERR_GENERIC);
	ERRORCASE(DDERR_NOTFOUND);
	ERRORCASE(DDERR_SURFACELOST);
	default:message("showerror: unknown ddraw error: %s:%d",func,error);break;
	}
}

int CDX7Video::Init(HWND wnd,HINSTANCE inst,int w,int h,int b)
{
typedef struct mode_s
	{
	u32 x,y;
	} mode_t;
mode_t modes[] =
	{
	{256,240},
	{256,256},
	{320,240},
	{640,480},
	{800,600}
	};

width = w;
height = h;
bpp = b;
hwnd = wnd;
hinstance = inst;
lpDD = 0;
lpDDSPrimary = 0;
lpDDSBack = 0;
renderstart = 0;
memset(&ddsd,0,sizeof(ddsd));
if(DirectDrawCreateEx(0,(VOID**)&lpDD,IID_IDirectDraw7,0))
	{
	message("dx7video_c::init: error: DirectDrawCreateEx\n");
	return(1);
	}
if(bpp == 0) //windowed mode
	{
	if(lpDD->SetCooperativeLevel(hwnd,DDSCL_NORMAL) != DD_OK)
		{
		message("dx7video_c::init: error: SetCooperativeLevel\n");
		Kill();
		return(1);
		}
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	if(lpDD->CreateSurface(&ddsd,&lpDDSPrimary,0) != DD_OK)
		{
		message("dx7video_c::init: error: CreateSurface\n");
		Kill();
		return(1);
		}
	if(lpDD->CreateClipper(0,&lpClipper, NULL) != DD_OK)
		{
		message("dx7video_c::init: error: CreateClipper\n");
		Kill();
		return(1);
		}
	lpClipper->SetHWnd(0,hwnd);
	lpDDSPrimary->SetClipper(lpClipper);
	lpClipper->Release();
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
	ddsd.dwWidth = w;
	ddsd.dwHeight = h;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	if(lpDD->CreateSurface(&ddsd,&lpDDSBack,0) != DD_OK)
		{
		message("dx7video_c::init: error: CreateSurface\n");
		Kill();
		return(1);
		}
	ddsd.dwSize = sizeof(ddsd);
	if(lpDDSBack->GetSurfaceDesc(&ddsd) != DD_OK)
		{
		message("dx7video_c::init: error: GetSurfaceDesc\n");
		Kill();
		return(1);
		}
	bpp = (u8)ddsd.ddpfPixelFormat.dwRGBBitCount;
	if(ddsd.ddpfPixelFormat.dwRBitMask == 0xf800 && bpp == 16)
		bpp++;
	switch(bpp & ~1)
		{
		case 8:screen8 = new u8[width * height];linesize = width;break;
		case 16:screen16 = new u16[width * height];linesize = width * 2;break;
		case 32:screen32 = new u32[width * height];linesize = width * 4;break;
		default:Kill();return(1);
		}
	Resize();
	mode = 0;
	lockflags = DDLOCK_WAIT;
	renderstart = 0;
	message("dx7video_c::init: inited window mode, %d bpp\n",bpp);
	}
else //fullscreen
	{
	DDSCAPS2 dc;
	DDBLTFX dbf;
	int i,ret;
u32 swidth,sheight;

	if((ret = DirectDrawCreateEx(0,(VOID**)&lpDD,IID_IDirectDraw7,0)) != DD_OK)
		{
		ShowError(ret,"DirectDrawCreateEx");
		Kill();
		return(1);
		}
	if((ret = lpDD->SetCooperativeLevel(hwnd,DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN)) != DD_OK)
		{
		ShowError(ret,"SetCooperativeLevel");
		Kill();
		return(1);
		}
	i = 0;
	ret = ~DD_OK;
	i = 3; //start at 640x480
	while(ret != DD_OK && i < 5)
		{
		swidth = modes[i].x;
		sheight = modes[i].y;
		i++;
		if((ret = lpDD->SetDisplayMode(swidth,sheight,bpp,0,0)) != DD_OK)
			ShowError(ret,"SetDisplayMode");
		}
	message("dx7video_c::init: using video mode %dx%d\n",swidth,sheight);
	memset(&ddsd,0,sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1;
	if((ret = lpDD->CreateSurface(&ddsd,&lpDDSPrimary,0)) != DD_OK)
		{
		ShowError(ret,"CreateSurface");
		Kill();
		return(1);
		}
	memset(&dc,0,sizeof(DDSCAPS2));
	dc.dwCaps = DDSCAPS_BACKBUFFER;
	if((ret = lpDDSPrimary->GetAttachedSurface(&dc,&lpDDSBack)) != DD_OK)
		{
		ShowError(ret,"GetAttachedSurface");
		Kill();
		return(1);
		}
	if((ret = lpDDSBack->GetSurfaceDesc(&ddsd)) != DD_OK)
		{
		ShowError(ret,"GetSurfaceDesc");
		Kill();
		return(1);
		}
	Resize();
	if(ddsd.ddpfPixelFormat.dwRBitMask == 0xf800 && bpp == 16)
		bpp++;
	if(bpp & 8)
		{
		message("no 8bit support\n");
		}
	memset(&dbf,0,sizeof(DDBLTFX));
	dbf.dwSize = sizeof(DDBLTFX);
	dbf.dwFillColor = 0;
	lpDDSBack->Blt(0,0,0,DDBLT_WAIT|DDBLT_COLORFILL,&dbf);
	lpDDSPrimary->Blt(0,0,0,DDBLT_WAIT|DDBLT_COLORFILL,&dbf);
	if((ret = lpDDSBack->Lock(0,&ddsd,DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT,0)) == DD_OK)
		{
		int xpos = 0;
		int ypos = 0;

		if(swidth >= (u32)width)
			xpos = ((swidth / 2) - (width / 2)) * (bpp / 8);
		if(sheight >= (u32)height)
			ypos = (sheight / 2) - (height / 2);
/*		if(bpp & 32 && swidth >= 640 && sheight >= 480 && effects)
			{
			xpos = ((swidth / 2) - ((width*2) / 2)) * (bpp / 8);
			ypos = (sheight / 2) - ((height*2) / 2);
			}*/
		renderstart = xpos + (ypos * ddsd.lPitch);
		lpDDSBack->Unlock(0);
		linesize = ddsd.lPitch;
		}
	else
		{
		ShowError(ret,"Lock");
		Kill();
		return(1);
		}
	mode = 1;
	lockflags = DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT;
	message("dx7video_c::init: inited fullscreen mode, %d bpp\n",bpp);
	}
return(0);
}

void CDX7Video::Kill()
{
if(lpDD != 0)
	{
	if(mode == 0)
		{
		switch(bpp & ~1)
			{
			case 8:delete[] screen8;break;
			case 16:delete[] screen16;break;
			case 32:delete[] screen32;break;
			}
		}
	if(mode == 1)
		lpDD->RestoreDisplayMode();
	lpDD->SetCooperativeLevel(hwnd,DDSCL_NORMAL);
	if(lpDDSBack != 0)
		{
		lpDDSBack->Release();
		lpDDSBack = 0;
		}
	if(lpDDSPrimary != 0)
		{
		lpDDSPrimary->Release();
		lpDDSPrimary = 0;
		}
	lpDD->Release();
	lpDD = 0;
	}
}

void CDX7Video::Resize()
{
RECT rectClient,rectWindow;
int w,h;
DDBLTFX dbf;

GetWindowRect(hwnd,&rectWindow);
GetClientRect(hwnd,&rectClient);
w = (rectWindow.right - rectWindow.left) - rectClient.right;
h = (rectWindow.bottom - rectWindow.top) - rectClient.bottom;
w /= 2;
GetWindowRect(hwnd,&rect);
rect.top += h - w;
rect.left += w;
rect.right -= w;
w++;
rect.bottom -= w;
if(lpDDSPrimary && lpDDSBack)
	{
	memset(&dbf,0,sizeof(DDBLTFX));
	dbf.dwSize = sizeof(DDBLTFX);
	dbf.dwFillColor = 0;
	lpDDSBack->Blt(0,0,0,DDBLT_WAIT|DDBLT_COLORFILL,&dbf);
	lpDDSPrimary->Blt(0,0,0,DDBLT_WAIT|DDBLT_COLORFILL,&dbf);
	}
}

void CDX7Video::StartBlit()
{
int ret;

if(lpDDSBack == 0)
	return;
if(mode == 1)
	{
	DDBLTFX dbf;

	dbf.dwSize = sizeof(DDBLTFX);
	dbf.dwFillColor = 0;
	lpDDSBack->Blt(0,0,0,DDBLT_WAIT|DDBLT_COLORFILL,&dbf);
	}
if((ret = lpDDSBack->Lock(0,&ddsd,lockflags,0)) == DD_OK)
	dest = ((u8*)ddsd.lpSurface) + renderstart;
else
	ShowError(ret,"CDX7Video::StartBlit");
}

void CDX7Video::StopBlit()
{
int ret = 0;

lpDDSBack->Unlock(0);
if(mode == 0)
	ret = lpDDSPrimary->Blt(&rect,lpDDSBack,0,DDBLT_WAIT,0);
else
	ret = lpDDSPrimary->Flip(0,0);
if(ret == DDERR_SURFACELOST)
	lpDDSPrimary->Restore();
}

void *CDX7Video::GetSurface()	{return(dest);}
int CDX7Video::GetPitch()		{return(linesize);}
