#include <windows.h>
#include <stdio.h>
#include <stdarg.h>

#include "main.h"
#include "dx7video.h"
#include "dxsound.h"
#include "wininput.h"

Cnes nes;							//nintendo nes components
Ccpu cpu;
Cppu ppu;
Capu apu;
CVideo *video = 0;				//video class
CInput *input = 0;				//input class
int quit = 0,running = 1;		//quit and running flags
HWND hwnd;							//main window handle
LARGE_INTEGER freq,c;			//frame time frequency/counter
__int64 frametime,last;			//frame time variables
char curdir[512];

void message(char *m,...)		//print a message
{
char buffer[256];					//buffer to store vsprintf'd message in
va_list argptr;					//argument data

va_start(argptr,m);				//get data
vsprintf(buffer,m,argptr);		//print to buffer
va_end(argptr);					//done!
printf(buffer);					//now display it
}

__inline void checkmessages()	//check for pending windows messages
{
MSG msg;

while(PeekMessage(&msg,0,0,0,PM_REMOVE))
	{
	TranslateMessage(&msg);
	DispatchMessage(&msg);
	}
}

int main(int argc, char *argv[])
{
HINSTANCE hinst = GetModuleHandle(0);
int w = 256,h = 240;

GetCurrentDirectory(512,curdir);
nes.cpu = &cpu;								//setup component pointers
nes.ppu = &ppu;
nes.apu = &apu;
nes.ppu->cpu = &cpu;
nes.cpu->debug = 0;							//debug level, mostly unused now...
QueryPerformanceFrequency(&freq);		//get counter frequency
QueryPerformanceCounter(&c);				//get counter data
last = c.QuadPart;
frametime = freq.QuadPart / 60;			//60 fps
if(nes.LoadROM(argv[1]) == 0)					//try to load the rom...
	nes.Reset();									//...and if success reset the nintendo
if((hwnd = createmainwindow(w,h)) == 0)	//try to create main window
	{
	message("error creating main window\n");
	return(1);
	}
video = new CDX7Video();
input = new CWinInput();
if(video->Init(hwnd,hinst,w,h) != 0)		//initialize video
	{
	destroymainwindow();
	delete video;
	message("error initializing directx video\n");
	return(1);
	}
if(input->Init(hwnd,hinst) != 0)				//initialize input
	{
	video->Kill();
	destroymainwindow();
	delete video;
	delete input;
	message("error initializing directx video\n");
	return(1);
	}
while(quit == 0)									//loop until quit != 0
	{
	if(running && nes.prgsize)					//if the nintendo is running...
		nes.Frame();								//...execute one frame
	for(;;)											//lock to 60 fps
		{
		QueryPerformanceCounter(&c);
		if((c.QuadPart - last) >= frametime)
			break;
		checkmessages();
		}
	QueryPerformanceCounter(&c);
	if((nes.ppu->frames & 0x1f) == 0)		//show fps
		{
		__int64 t = c.QuadPart - last;
		char fpsstr[128];

		if(t)
			sprintf(fpsstr,"amphetamines - %5.1f fps",(float)freq.QuadPart / t);
		SetWindowText(hwnd,fpsstr);
		}
	last = c.QuadPart;
	checkmessages();								//check for windows messages
	}
video->Kill();										//kill the video unit
input->Kill();
delete video;
delete input;
destroymainwindow();
return 0;											//return success
}
