class CMMC1:public CMapper
	{
	public:
		u8 Regs[4];					//mmc registers
		u8 Latch,LatchPos;		//latch and latch position counter
		int Last;					//last register written to
	public:
		CMMC1()
			{
			}
		void Reset()
			{
			SetWriteHandled(0x8,0xF);
			nes.SetSizeSRAM(8);
			if(nes.chrsize == 0)
				nes.SetSizeVRAM(8);
			Regs[0] = 0xC;
			Regs[1] = Regs[2] = Regs[3] = 0;
			Latch = LatchPos = 0;
			Sync();
			}
		virtual void Sync()
			{
			SyncPRG();
			if(nes.chrsize == 0)
				SyncVRAM();
			else
				SyncCHR();
			SyncMirroring();
			}
		void SyncPRG()
			{
			nes.SetPageCPU16(0x8,GetLowPRG());
			nes.SetPageCPU16(0xC,GetHighPRG());
			}
		void SyncCHR()
			{
			nes.SetPagePPU4(0,GetLowCHR());
			nes.SetPagePPU4(4,GetHighCHR());
			}
		void SyncVRAM()
			{
			nes.SetPageVRAM4(0,GetLowCHR());
			nes.SetPageVRAM4(4,GetHighCHR());
			}
		void SyncMirroring()
			{
			switch(Regs[0] & 3)
				{
				case 0:nes.ppu->SetMirroring(MIRROR_1L);break;
				case 1:nes.ppu->SetMirroring(MIRROR_1H);break;
				case 2:nes.ppu->SetMirroring(1);break;
				case 3:nes.ppu->SetMirroring(0);break;
				}
			}
		int GetLowPRG()
			{
			if(Regs[0] & 8)
				{
				if(Regs[0] & 4)
					return(Regs[3] & 0xF);
				else
					return(0);
				}
			else
				return(Regs[3] & 0xE);
			}
		int GetHighPRG()
			{
			if(Regs[0] & 8)
				{
				if(Regs[0] & 4)
					return(0xF);
				else
					return(Regs[3] & 0xF);
				}
			else
				return((Regs[3] & 0xE) | 1);
			}
		int GetLowCHR()
			{
			if(Regs[0] & 0x10)
				return(Regs[1] & 0x1F);
			else
				return(Regs[1] & 0x1E);
			}
		int GetHighCHR()
			{
			if(Regs[0] & 0x10)
				return(Regs[2] & 0x1F);
			else
				return((Regs[1] & 0x1E) | 1);
			}
		void Write(u32 addr,u8 data)
			{
			int t = ((addr >> 12) >> 1) & 3;

			if(data & 0x80)
				{
				Latch = LatchPos = 0;
				Regs[0] |= 0xC;
				return;
				}
			if(t != Last)
				Latch = LatchPos = 0;
			Last = t;
			Latch |= (data & 1) << LatchPos++;
			if(LatchPos == 5)
				{
				Regs[t] = Latch & 0x1F;
				Latch = LatchPos = 0;
				Sync();
				}
			}
	};
