class CMMC3:public CMapper
	{
	public:
		u8 PRGBank[4],CHRBank[8];
		u8 IRQEnabled,IRQCount,IRQLatch;
		u8 Command;
		u8 Mirroring;
		u8 SRAMEnabled;
		int Line;
	public:
		CMMC3()
			{
			}
		void Reset()
			{
			SetWriteHandled(0x8,0xF);
			nes.ppu->cycle = MapperCycle;
			nes.SetSizeSRAM(8);
			if(nes.chrsize == 0)
				nes.SetSizeVRAM(8);
			PRGBank[0] = 0x3C;
			PRGBank[1] = 0x3D;
			PRGBank[2] = 0x3E;
			PRGBank[3] = 0x3F;
			CHRBank[0] = 0;
			CHRBank[1] = 1;
			CHRBank[2] = 2;
			CHRBank[3] = 3;
			CHRBank[4] = 4;
			CHRBank[5] = 5;
			CHRBank[6] = 6;
			CHRBank[7] = 7;
			IRQEnabled = 0;
			IRQCount = 0;
			IRQLatch = 0;
			Command = 0;
			Line = 0;
			Sync();			
			}
		void Scanline(int line)
			{
			if(line == -1)
				IRQCount = IRQLatch;
			else if(IRQEnabled && (line < 240) && (nes.ppu->control1 & 0x18))
				{
				if(IRQCount-- == 0)
					{
					IRQCount = IRQLatch;
					nes.cpu->IRQ();
					IRQEnabled = 0;
					}
				}
			}
//		u32 lastppuaddr; 
		void Cycle(int pixel)
			{
/*			if(nes.ppu->ppuaddr & 0x1000 && (lastppuaddr & 0x1000) == 0)
				{
				message("counter clocked, pixel = %d, scanline = %d ",pixel,nes.ppu->scanline);
				message("ppuaddr = $%04X ",nes.ppu->ppuaddr);
				message("last = $%04X\n",lastppuaddr);
				if(IRQCount == 0)
					IRQCount = IRQLatch;
				else
					IRQCount--;
				if(IRQEnabled && IRQCount == 0)
					{
					IRQCount = IRQLatch;
					nes.cpu->IRQ();
					IRQEnabled = 0;
					}
				}
			lastppuaddr = nes.ppu->ppuaddr;*/
			if(nes.ppu->pixel == 0)
				{
				if(nes.ppu->scanline >= 20)
					Scanline(nes.ppu->scanline - 21);
				}
			}
		virtual void Sync()
			{
			SyncPRG();
			if(nes.chrsize)
				SyncCHR();
			else
				SyncVRAM();
			SyncMirroring();
			}
		void SyncPRG()
			{
			nes.SetPageCPU8(0x8,GetPRG(0));
			nes.SetPageCPU8(0xA,GetPRG(1));
			nes.SetPageCPU8(0xC,GetPRG(2));
			nes.SetPageCPU8(0xE,GetPRG(3));
			}
		void SyncCHR()
			{
			int i;
			
			for(i=0;i<8;i++)
				nes.SetPagePPU1(i,GetCHR(i));
			}
		void SyncVRAM()
			{
			int i;
			
			for(i=0;i<8;i++)
				nes.SetPageVRAM1(i,GetCHR(i));
			}
		void SyncMirroring()
			{
			nes.ppu->SetMirroring(~Mirroring & 1);
			}
		virtual int GetPRG(int bank)
			{
			if(bank & 1)
				return(PRGBank[bank]);
			return(PRGBank[bank ^ ((Command & 0x40) >> 5)]);
			}
		virtual int GetCHR(int bank)
			{
			return(CHRBank[bank ^ ((Command & 0x80) >> 5)]);
			}
		void Write(u32 addr,u8 data)
			{
			switch(addr & 0xE001)
				{
				case 0x8000:
					Command = data;
					break;
				case 0x8001:
					switch(Command & 7)
						{
						case 0:
							CHRBank[0] = data & 0xFE;
							CHRBank[1] = (data & 0xFE) | 1;
							break;
						case 1:
							CHRBank[2] = data & 0xFE;
							CHRBank[3] = (data & 0xFE) | 1;
							break;
						case 2:	CHRBank[4] = data;		break;
						case 3:	CHRBank[5] = data;		break;
						case 4:	CHRBank[6] = data;		break;
						case 5:	CHRBank[7] = data;		break;
						case 6:	PRGBank[0] = data & 0x3F;	break;
						case 7:	PRGBank[1] = data & 0x3F;	break;
						}
					break;
				case 0xA000:Mirroring = data;Sync();break;
				case 0xA001:SRAMEnabled = data;break;
				case 0xC000:IRQLatch = data;break;
				case 0xC001:IRQCount = 0;break;
				case 0xE000:IRQEnabled = 0;break;
				case 0xE001:IRQEnabled = 1;break;
				}
			Sync();
			}
	};
