#include <windows.h>
#include <stdio.h>

#include "nes.h"
#include "main.h"
#include "mappers.h"

Cnes::Cnes()
{
rom = chrrom = 0;
prgsize = chrsize = 0;
mask = 0;
cpu = 0;
ppu = 0;
apu = 0;
}

int Cnes::LoadROM(char *filename)
{
int m,ret;
FILE *fp;

if(filename == 0)
	{
	message("Cnes::LoadROM: error: filename = 0\n");
	return(2);
	}
if((fp = fopen(filename,"rb")) == 0)
	{
	message("cant open %s\n",filename);
	return(1);
	}
if(rom)
	free(rom);
if(chrrom)
	free(chrrom);
rom = chrrom = 0;
fread(&header,1,16,fp);
prgsize = header[4] * 0x4000;
chrsize = header[5] * 0x2000;
message("%dk prg, %dk chr\n",prgsize/1024,chrsize/1024);
rom = (u8*)malloc(prgsize);
fread(rom,1,prgsize,fp);
if(chrsize)
	{
	chrrom = (u8*)malloc(chrsize);
	fread(chrrom,1,chrsize,fp);
	}
mask = prgsize - 1;
m = (header[6] >> 4) | (header[7] & 0xF0);
fclose(fp);
ret = MapperInit(m);
if(ret != 0)
	{
	free(rom);
	if(chrrom)
		free(chrrom);
	rom = chrrom = 0;
	prgsize = chrsize = 0;
	}
return(ret);
}

int cpustep;

__inline void Cnes::Step()
{
if(cpustep == 0)
	{
	cpu->Step();
//	apu->Step();
	ppu->Step();
	cpustep++;
	}
else if(cpustep == 1)
	{
	ppu->Step();
	cpustep++;
	}
else if(cpustep == 2)
	{
//	cpu->Step();
	ppu->Step();
	cpustep = 0;
	}
}

void Cnes::Execute(int n)
{
while(n && cpu->badopcode == 0)
	{
   Step();
	n--;
   }
}

void Cnes::Frame()
{
video->StartBlit();									//start directx blit (fetch screen ptr, etc)
nes.ppu->scr32 = (u32*)video->GetSurface();	//get pointer to actual screen to draw on
nes.ppu->pitch = video->GetPitch();				//get size of each line in bytes
if(ppu->framecycles == 0)
	Step();
while(ppu->framecycles != 0)
	Step();
video->StopBlit();									//tell dx7 class to end frame and draw it
}

void Cnes::Reset()
{
if(prgsize == 0)
	{
	message("error:  cannot reset without any loaded data\n");
	return;
	}
cpu->SetPPU(ppu);									//set active ppu
cpu->SetAPU(apu);									//set active apu
memset(ram,0,0x800);								//zero the ram (hard reset style)
memset(vram,0,0x4000);							//zero the vram
memset(sram,0,0x4000);							//zero the sram
cpu->ramptr[0] = cpu->romptr[0] = ram;		//set the page ptrs to ram and its three mirrors
cpu->ramptr[1] = cpu->romptr[1] = ram;
cpu->ramptr[2] = cpu->romptr[2] = ram;
cpu->ramptr[3] = cpu->romptr[3] = ram;
ppu->Reset();										//ppu must be reset before cpu
ppu->GeneratePalettes();
ppu->SetMirroring(header[6] & 1);
if(header[6] & 4)
	ppu->SetMirroring(4);
MapperReset();
cpu->Reset();
cpustep = 0;
}

int Cnes::SaveState(char *fn)
{
return(0);
}

int Cnes::LoadState(char *fn)
{
return(0);
}

void Cnes::SetSizeSRAM(int s)
{
cpu->ramptr[12] = cpu->romptr[12] = sram;
cpu->ramptr[13] = cpu->romptr[13] = sram + 0x800;
cpu->ramptr[14] = cpu->romptr[14] = sram + 0x1000;
cpu->ramptr[15] = cpu->romptr[15] = sram + 0x1800;
}

void Cnes::SetSizeVRAM(int s)
{
message("vram size set to %dk\n",s);
SetPageVRAM8(0,0);
}

void Cnes::SetPageCPU8(int p,int d)
{
int dd = (d * 0x800 * 4) & mask;

if(d == -1)
	dd = nes.prgsize - 0x2000;
//message("Cnes::SetPageCPU8: setting 8k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+2)*2,dd,d);
p *= 2;
cpu->romptr[p] = rom + dd;
cpu->romptr[p+1] = rom + dd + 0x800;
cpu->romptr[p+2] = rom + dd + 0x1000;
cpu->romptr[p+3] = rom + dd + 0x1800;
}
void Cnes::SetPageCPU16(int p,int d)
{
int dd = (d * 0x800 * 8) & mask;

if(d == -1)
	dd = nes.prgsize - 0x4000;
//message("Cnes::SetPageCPU16: setting 16k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+4)*2,dd,d);
p *= 2;
cpu->romptr[p] = rom + dd;
cpu->romptr[p+1] = rom + dd + 0x800;
cpu->romptr[p+2] = rom + dd + 0x1000;
cpu->romptr[p+3] = rom + dd + 0x1800;
cpu->romptr[p+4] = rom + dd + 0x2000;
cpu->romptr[p+5] = rom + dd + 0x2800;
cpu->romptr[p+6] = rom + dd + 0x3000;
cpu->romptr[p+7] = rom + dd + 0x3800;
}
void Cnes::SetPageCPU32(int p,int d)
{
int dd = (d * 0x800 * 16) & mask;

if(d == -1)
	dd = nes.prgsize - 0x8000;
//message("Cnes::SetPageCPU32: setting 32k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+8)*2,dd,d);
p *= 2;
cpu->romptr[p] = rom + dd;
cpu->romptr[p+1] = rom + dd + 0x800;
cpu->romptr[p+2] = rom + dd + 0x1000;
cpu->romptr[p+3] = rom + dd + 0x1800;
cpu->romptr[p+4] = rom + dd + 0x2000;
cpu->romptr[p+5] = rom + dd + 0x2800;
cpu->romptr[p+6] = rom + dd + 0x3000;
cpu->romptr[p+7] = rom + dd + 0x3800;
cpu->romptr[p+8] = rom + dd + 0x4000;
cpu->romptr[p+9] = rom + dd + 0x4800;
cpu->romptr[p+10] = rom + dd + 0x5000;
cpu->romptr[p+11] = rom + dd + 0x5800;
cpu->romptr[p+12] = rom + dd + 0x6000;
cpu->romptr[p+13] = rom + dd + 0x6800;
cpu->romptr[p+14] = rom + dd + 0x7000;
cpu->romptr[p+15] = rom + dd + 0x7800;
}

void Cnes::SetPagePPU1(int p,int d)
{
int dd = (d * 0x400) & (chrsize-1);

ppu->readptr[p] = chrrom + dd;
//message("Cnes::SetPagePPU1: setting 1k page %d to $%X (d=$%02X)\n",p,dd,d);
}

void Cnes::SetPagePPU4(int p,int d)
{
int dd = (d * 0x400 * 4) & (chrsize-1);

ppu->readptr[p] = chrrom + dd;
ppu->readptr[p+1] = chrrom + dd + 0x400;
ppu->readptr[p+2] = chrrom + dd + 0x800;
ppu->readptr[p+3] = chrrom + dd + 0xC00;
//message("Cnes::SetPagePPU4: setting 4k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+2)*2,dd,d);
}

void Cnes::SetPagePPU8(int p,int d)
{
int dd = (d * 0x400 * 8) & (chrsize-1);

ppu->readptr[p] = chrrom + dd;
ppu->readptr[p+1] = chrrom + dd + 0x400;
ppu->readptr[p+2] = chrrom + dd + 0x800;
ppu->readptr[p+3] = chrrom + dd + 0xC00;
ppu->readptr[p+4] = chrrom + dd + 0x1000;
ppu->readptr[p+5] = chrrom + dd + 0x1400;
ppu->readptr[p+6] = chrrom + dd + 0x1800;
ppu->readptr[p+7] = chrrom + dd + 0x1C00;
//message("Cnes::SetPagePPU8: setting 8k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+4)*2,dd,d);
}

void Cnes::SetPageVRAM1(int p,int d)
{
int dd = (d * 0x400) & (0x4000-1);

ppu->readptr[p] = ppu->writeptr[p] = vram + dd;
}

void Cnes::SetPageVRAM4(int p,int d)
{
int dd = (d * 0x400 * 4) & (0x4000-1);

ppu->readptr[p] = ppu->writeptr[p] = vram + dd;
ppu->readptr[p+1] = ppu->writeptr[p+1] = vram + dd + 0x400;
ppu->readptr[p+2] = ppu->writeptr[p+2] = vram + dd + 0x800;
ppu->readptr[p+3] = ppu->writeptr[p+3] = vram + dd + 0xC00;
//message("Cnes::SetPageVRAM4: setting 4k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+2)*2,dd,d);
}

void Cnes::SetPageVRAM8(int p,int d)
{
int dd = (d * 0x400 * 8) & (0x4000-1);

ppu->readptr[p] = ppu->writeptr[p] = vram + dd;
ppu->readptr[p+1] = ppu->writeptr[p+1] = vram + dd + 0x400;
ppu->readptr[p+2] = ppu->writeptr[p+2] = vram + dd + 0x800;
ppu->readptr[p+3] = ppu->writeptr[p+3] = vram + dd + 0xC00;
ppu->readptr[p+4] = ppu->writeptr[p+4] = vram + dd + 0x1000;
ppu->readptr[p+5] = ppu->writeptr[p+5] = vram + dd + 0x1400;
ppu->readptr[p+6] = ppu->writeptr[p+6] = vram + dd + 0x1800;
ppu->readptr[p+7] = ppu->writeptr[p+7] = vram + dd + 0x1C00;
//message("Cnes::SetPageVRAM4: setting 4k page %d (%d-%d) to $%X (d=$%02X)\n",p,p*2,(p+2)*2,dd,d);
}
