#include <memory.h>

#include "ppu.h"

typedef struct palentry_s
	{
	u8 r,g,b;
	} palentry_t;

static palentry_t ronipalette[64] = //palette by roni, the default
	{
	{0x75,0x75,0x75},{0x27,0x1b,0x8f},{0x00,0x00,0xab},{0x47,0x00,0x9f}, //0x00 - 0x03
	{0x8f,0x00,0x77},{0xab,0x00,0x13},{0xa7,0x00,0x00},{0x7f,0x0b,0x00}, //0x04 - 0x07
	{0x43,0x2f,0x00},{0x00,0x47,0x00},{0x00,0x51,0x00},{0x00,0x3f,0x17}, //0x08 - 0x0b
	{0x1b,0x3f,0x5f},{0x00,0x00,0x00},{0x00,0x00,0x00},{0x00,0x00,0x00}, //0x0c - 0x0f
	{0xbc,0xbc,0xbc},{0x00,0x73,0xef},{0x23,0x3b,0xef},{0x83,0x00,0xf3}, //0x10 - 0x13
	{0xbf,0x00,0xbf},{0xe7,0x00,0x5b},{0xdb,0x2b,0x00},{0xcb,0x4f,0x0f}, //0x14 - 0x17
	{0x8b,0x73,0x00},{0x00,0x97,0x00},{0x00,0xab,0x00},{0x00,0x93,0x3b}, //0x18 - 0x1b
	{0x00,0x83,0x8b},{0x00,0x00,0x00},{0x00,0x00,0x00},{0x00,0x00,0x00}, //0x1c - 0x1f
	{0xff,0xff,0xff},{0x3f,0xbf,0xff},{0x5f,0x97,0xff},{0xa7,0x8b,0xfd}, //0x20 - 0x23
	{0xf7,0x7b,0xff},{0xff,0x77,0xb7},{0xff,0x77,0x63},{0xff,0x9b,0x3b}, //0x24 - 0x27
	{0xf3,0xbf,0x3f},{0x83,0xd3,0x13},{0x4f,0xdf,0x4b},{0x58,0xf8,0x98}, //0x28 - 0x2b
	{0x00,0xeb,0xdb},{0x00,0x00,0x00},{0x00,0x00,0x00},{0x00,0x00,0x00}, //0x2c - 0x2f
	{0xff,0xff,0xff},{0xab,0xe7,0xff},{0xc7,0xd7,0xff},{0xd7,0xcb,0xff}, //0x30 - 0x33
	{0xff,0xc7,0xff},{0xff,0xc7,0xdb},{0xff,0xbf,0xb3},{0xff,0xdb,0xab}, //0x34 - 0x37
	{0xe1,0xe7,0xa3},{0xe3,0xff,0xa3},{0xab,0xf3,0xbf},{0xb3,0xff,0xcf}, //0x38 - 0x3b
	{0x9f,0xff,0xf3},{0x00,0x00,0x00},{0x00,0x00,0x00},{0x00,0x00,0x00}  //0x3c - 0x3f
	};

void message(char *m,...);

Cppu::Cppu()
{
SetMirroring(-1);
cycle = 0;
evenodd = 0;	//even frame
}

void Cppu::GeneratePalettes()
{
palentry_t *p = ronipalette;
int i;
int r,g,b;

for(i=0;i<64;i++)
	{
	pal32[0][i] =	pal32[0][i+0x40] =	pal32[0][i+0x80] =
	pal32[0][i+0xC0] =  p[i].b | (p[i].g << 8) | (p[i].r << 16);
	r = (int)((float)p[i].r * 1.239f);
	g = (int)((float)p[i].g * 0.915f);
	b = (int)((float)p[i].b * 0.743f);
	pal32[1][i] = pal32[1][i+0x40] = pal32[1][i+0x80] = pal32[1][i+0xC0] =  b | (g<<8) | (r<<16);
	r = (int)((float)p[i].r * 0.794f);
	g = (int)((float)p[i].g * 1.086f);
	b = (int)((float)p[i].b * 0.882f);
	pal32[2][i] = pal32[2][i+0x40] = pal32[2][i+0x80] = pal32[2][i+0xC0] =  b | (g<<8) | (r<<16);
	r = (int)((float)p[i].r * 1.019f);
	g = (int)((float)p[i].g * 0.980f);
	b = (int)((float)p[i].b * 0.653f);
	pal32[3][i] = pal32[3][i+0x40] = pal32[3][i+0x80] = pal32[3][i+0xC0] =  b | (g<<8) | (r<<16);
	r = (int)((float)p[i].r * 0.905f);
	g = (int)((float)p[i].g * 1.026f);
	b = (int)((float)p[i].b * 1.277f);
	pal32[4][i] = pal32[4][i+0x40] = pal32[4][i+0x80] = pal32[4][i+0xC0] =  b | (g<<8) | (r<<16);
	r = (int)((float)p[i].r * 0.979f);
	g = (int)((float)p[i].g * 0.908f);
	b = (int)((float)p[i].b * 1.023f);
	pal32[5][i] = pal32[5][i+0x40] = pal32[5][i+0x80] = pal32[5][i+0xC0] =  b | (g<<8) | (r<<16);
	r = (int)((float)p[i].r * 0.979f);
	g = (int)((float)p[i].g * 0.987f);
	b = (int)((float)p[i].b * 1.001f);
	pal32[6][i] = pal32[6][i+0x40] = pal32[6][i+0x80] = pal32[6][i+0xC0] =  b | (g<<8) | (r<<16);
	r = (int)((float)p[i].r * 0.750f);
	g = (int)((float)p[i].g * 0.750f);
	b = (int)((float)p[i].b * 0.750f);
	pal32[7][i] = pal32[7][i+0x40] = pal32[7][i+0x80] = pal32[7][i+0xC0] =  b | (g<<8) | (r<<16);
	}
}

void Cppu::Reset()
{
int i;

message("ppu reset\n");
status = 0x80;
pixel = 0;
scanline = 0;
frames = 0;
toggle = 0;
spr0hit = 0;
framecycles = 0;
for(i=0;i<0x1000;i++)
	ntram[i] = 0;
for(i=0;i<32;i++)
	palram[i] = 0x0F;
for(i=0;i<16;i++)
	{
	readptr[i] = 0;
	writeptr[i] = 0;
	readfunc[i] = 0;
	writefunc[i] = 0;
	}
}

void Cppu::SetNT(int nt,int bank)
{
int n;

nt &= 3;
n = nt | 8;
readptr[n] = writeptr[n] = readptr[n+4] = writeptr[n+4] = ntram + (0x400 * nt);
}

void Cppu::SetMirroring(int m)
{
switch(m)
	{
	default:
	case 0:		//horizontal
		SetNT(0,0);
		SetNT(1,0);
		SetNT(2,1);
		SetNT(3,1);
		break;
	case 1:		//vertical
		SetNT(0,0);
		SetNT(1,1);
		SetNT(2,0);
		SetNT(3,1);
		break;
	case 4:		//four screen
		SetNT(0,0);
		SetNT(1,1);
		SetNT(2,2);
		SetNT(3,3);
		break;
	case MIRROR_1L:		//one screen low
		SetNT(0,0);
		SetNT(1,0);
		SetNT(2,0);
		SetNT(3,0);
		break;
	case MIRROR_1H:		//one screen high
		SetNT(0,1);
		SetNT(1,1);
		SetNT(2,1);
		SetNT(3,1);
		break;
	}
}

u8 Cppu::MemRead(u32 addr)
{
int b = (addr & 0x3FFF) >> 10;

if(readfunc[b])
	return(readfunc[b](addr));
if(readptr[b])
	return(readptr[b][addr & 0x3FF]);
//message("read from unmapped ppu address $%04X (scanline=%d, pixel=%d)\n",addr,scanline,pixel);
return(0);
}

void Cppu::MemWrite(u32 addr,u8 data)
{
int b = (addr & 0x3FFF) >> 10;

if(writefunc[b])
	{
	writefunc[b](addr,data);
	return;
	}
if(writeptr[b])
	writeptr[b][addr & 0x3FF] = data;
//else
//	message("write to unmapped ppu address $%04X (%d)\n",addr,b);
}

#define PIPECOPY(n1,n2)						\
	pipeline[n1].nt = pipeline[n2].nt;		\
	pipeline[n1].attr = pipeline[n2].attr;	\
	pipeline[n1].p[0] = pipeline[n2].p[0];	\
	pipeline[n1].p[1] = pipeline[n2].p[1];

void Cppu::GetSprites()
{
int i,h;
u8 *s;

h = 8;
h += (control1 & 0x20) >> 2;					//detect 16 pixel high sprites
for(sprinrange=0,i=0;i<64;i++)				//loop thru all 64 sprites
	{
	int ss,n,s3;
	u8 s2;

	s = &sprmem[i * 4];
	s2 = *s + 1;
	ss = scanline - 21;
	if(ss >= s2 && ss < (s2 + h))				//is sprite is visible on this line?
		{
		if(sprinrange < 8)						//already found 8 sprites?
			{
			sprtemp[sprinrange].attr = s[2];	//copy sprite data to temp memory
			sprtemp[sprinrange].x = s[3] + finex;
			sprtemp[sprinrange].tile = s[1];
			sprtemp[sprinrange].tmpx = 0;
			sprtemp[sprinrange].index = i;
			if(h == 8)
				n = (s[1] * 16) | ((control1 & 8) << 9);
			else
				n = ((s[1] & ~1) * 16) | ((s[1] & 1) << 12);
			s3 = ss - s2;
			if(sprtemp[sprinrange].attr & 0x80)	//handle vflip
				{
				n += 7;
				n -= s3;
				n += (control2 & 0x20) >> 1;
				n -= s3 & 8;
				}
			else
				{
				n += s3;
				n += s3 & 8;
				}
			sprtemp[sprinrange].p[0] = readptr[(n >> 10) & 7][n & 0x3FF];
			n += 8;
			sprtemp[sprinrange].p[1] = readptr[(n >> 10) & 7][n & 0x3FF];
			sprinrange++;						//sprite is in range!
			}
		else
			status |= 0x20;						//8+ sprites found, set the flag
		}
	}
}

void Cppu::Step()
{
int i,h;

switch(pixel)
	{
	case 0:
		if(scanline == 20)
			status &= ~(0x40 | 0x20);	//clear spr0hit, spr8 flags
		if((scanline >= 21) && (scanline <= 261) && (control2 & 0x10))
			GetSprites();					//get sprites for this scanline into sprite temp memory
		break;
	case 252:
		if(scanline == 20)
			{
			status = 0;
			spraddr = 0;
			if(control2 & 0x18)
				ppuaddr = pputemp;
			}
		break;
	case 256:
		if(scanline >= 21 && scanline <= 261 && control2 & 0x18)
			{
			ppuaddr &= ~0x41F;
			ppuaddr |= pputemp & 0x41F;
			if((ppuaddr & 0x7000) == 0x7000)
				{
				int t = ppuaddr & 0x3E0;

				ppuaddr &= 0xFFF;
				if(t == 0x3A0)
					ppuaddr ^= 0xBA0;
				else if(t == 0x3E0)
					ppuaddr ^= 0x3E0;
				else
					ppuaddr += 0x20;
				}
			else
				ppuaddr += 0x1000;
			}
		break;

/*	1. Name table byte
	2. Attribute table byte
	3. Pattern table bitmap #0
	4. Pattern table bitmap #1 */
	//fetch name table byte
/*	case 0:*/case 8:	case 16:	case 24:	case 32:	case 40:	case 48:	case 56:
	case 64:	case 72:	case 80:	case 88:	case 96:	case 104:case 112:case 120:
	case 128:case 136:case 144:case 152:case 160:case 168:case 176:case 184:
	case 192:case 200:case 208:case 216:case 224:case 232:case 240:case 248:
		if(scanline >= 21 && scanline <= 261 && control2 & 0x18)
			{
			PIPECOPY(0,1);
			PIPECOPY(1,2);
			}
		break;
	case 1:	
		if(scanline == 0 && control1 & 0x80)							//ctrl1 requesting nmi?
			{
			cpu->NMI();													//do nmi
			//message("nmi happened on cycle %ld\n",framecycles);
			}
	case 9:  case 17: case 25: case 33: case 41: case 49: case 57:
	case 65: case 73: case 81: case 89: case 97: case 105:case 113:case 121:
	case 129:case 137:case 145:case 153:case 161:case 169:case 177:case 185:
	case 193:case 201:case 209:case 217:case 225:case 233:case 241:case 249:
		if(scanline >= 21 && scanline <= 261 && control2 & 0x18)
			{
			pipeline[2].nt = readptr[((ppuaddr >> 10) & 3) + 8][ppuaddr & 0x3FF];
			pipeline[2].chraddr = (pipeline[2].nt * 16) + ((control1 & 0x10) << 8);
			pipeline[2].chraddr += ppuaddr >> 12;
			}
		break;

	//fetch attribute byte and clock the HT counter
	case 3:	case 11: case 19: case 27:	case 35: case 43: case 51: case 59:
	case 67: case 75: case 83: case 91:	case 99: case 107:case 115:case 123:
	case 131:case 139:case 147:case 155:case 163:case 171:case 179:case 187:
	case 195:case 203:case 211:case 219:case 227:case 235:case 243:case 251:
		if(scanline >= 21 && scanline <= 261 && control2 & 0x18)
			{
			u8 a;
			u32 attraddr;

			attraddr = ppuaddr & 0x3C00;
			attraddr += 0x3c0 + ((ppuaddr >> 2) & 7) + (((ppuaddr >> (2 + 5)) & 7) << 3);
			a = readptr[((attraddr >> 10) & 3) + 8][attraddr & 0x3FF];
			pipeline[2].attr = ((a >> (((ppuaddr & 2) | (((ppuaddr >> 5) & 2) << 1)))) & 3) << 2;
/*	The first one, the horizontal scroll counter, consists of 6 bits, and is 
	made up by daisy-chaining the HT counter to the H counter. The HT counter is 
	then clocked every 8 pixel dot clocks (or every 8/3 CPU clock cycles). */
			if((ppuaddr & 0x1F) == 0x1F)				//see if HT counter creates carry
				ppuaddr ^= 0x41F;							//yes, clear lower 5 bits and toggle H counter
			else
				ppuaddr++;									//no, increment address
			}
		break;

	//fetch first pattern table byte
	case 5:	 case 13: case 21: case 29: case 37: case 45: case 53: case 61:
	case 69: case 77: case 85: case 93: case 101:case 109:case 117:case 125:
	case 133:case 141:case 149:case 157:case 165:case 173:case 181:case 189:
	case 197:case 205:case 213:case 221:case 229:case 237:case 245:case 253:
		if(scanline >= 21 && scanline <= 261 && control2 & 0x18)
			pipeline[2].p[0] = readptr[(pipeline[2].chraddr >> 10) & 7][pipeline[2].chraddr & 0x3FF];
		break;

	//fetch second pattern table byte
	case 7:	 case 15: case 23: case 31: case 39: case 47: case 55: case 63:
	case 71: case 79: case 87: case 95: case 103:case 111:case 119:case 127:
	case 135:case 143:case 151:case 159:case 167:case 175:case 183:case 191:
	case 199:case 207:case 215:case 223:case 231:case 239:case 247:case 255:
		if(scanline >= 21 && scanline <= 261 && control2 & 0x18)
			{
			int p = pipeline[2].chraddr + 8;

			pipeline[2].p[1] = readptr[(p >> 10) & 7][p & 0x3FF];
			}
		break;

	//start playfield pipeline
	case 321:case 329:
		if(scanline >= 20 && scanline <= 261 && control2 & 0x18)
			{
			pipeline[2].nt = readptr[((ppuaddr >> 10) & 3) + 8][ppuaddr & 0x3FF];
			pipeline[2].chraddr = (pipeline[2].nt * 16) + ((control1 & 0x10) << 8);
			pipeline[2].chraddr += ppuaddr >> 12;
			}
		break;
	case 323:case 331:
		if(scanline >= 20 && scanline <= 261 && control2 & 0x18)
			{
			u8 a;
			u32 attraddr;

			attraddr = ppuaddr & 0x3C00;
			attraddr += 0x3c0 + ((ppuaddr >> 2) & 7) + (((ppuaddr >> (2 + 5)) & 7) << 3);
			a = readptr[((attraddr >> 10) & 3) + 8][attraddr & 0x3FF];
			pipeline[2].attr = ((a >> (((ppuaddr & 2) | (((ppuaddr >> 5) & 2) << 1)))) & 3) << 2;
			if((ppuaddr & 0x1F) == 0x1F)				//see if HT counter creates carry
				ppuaddr ^= 0x41F;							//yes, clear lower 5 bits and toggle H counter
			else
				ppuaddr++;									//no, increment address
			}
		break;
	case 325:case 333:
		if(scanline >= 20 && scanline <= 261 && control2 & 0x18)
			pipeline[2].p[0] = readptr[(pipeline[2].chraddr >> 10) & 7][pipeline[2].chraddr & 0x3FF];
		break;
	case 327:
		if(scanline == 19)
			{
			//message("vint clear on cycle %ld\n",framecycles);
			status &= ~0x80;
			}
	case 335:
		if(scanline >= 20 && scanline <= 261 && control2 & 0x18)
			{
			int p = pipeline[2].chraddr + 8;

			pipeline[2].p[1] = readptr[(p >> 10) & 7][p & 0x3FF];
			}
		break;
	case 328:
		if(scanline == 20 && evenodd == 1 && control2 & 8)	//odd frame and bg is enabled
			{
			//is this correct?  doesnt act right...
			pixel++;							//execute two pixels here
			this->Step();					//call itself
			pixel--;
			framecycles--;					//include this cycle with previous cycle
			cpu->EatCycles(1);
			}
	case 336:
		if(scanline >= 20 && scanline <= 261 && control2 & 0x18)
			{
			PIPECOPY(0,1);
			PIPECOPY(1,2);
			}
	//wasted accesses
	case 337:case 338:case 339:
		//MemRead(ppuaddr & 0x3FFF);
		break;
	case 340:								//last pixel, wasted cycle
		if(scanline >= 21 && scanline <= 261)
			memcpy(scr32 + ((scanline-21)*pitch/4),linepixels + finex,(256-finex)*4);
//			memcpy(scr32 + ((scanline-21)*pitch/4),linepixels,256*4);
		if(scanline == 261)
			{
			if(evenodd == 0)
				evenodd = 1;
			else
				evenodd = 0;
			status |= 0x80;					//i assume vint is set on this cycle
			scanline = 0;
			spr0hit = 0;
			frames++;
			framecycles = 0;
			}
		else
			{
			framecycles++;
			scanline++;
			}
		pixel = 0;
		if(cycle)
			cycle(pixel);
		return;
	}
if(scanline >= 21 && scanline <= 261 && pixel < 256)	//a pixel is rendered every cycle
	{
	int d = scanline - 21;
	int n = 7 - (pixel & 7);
	u8 p = 0,sp = 0;

	h = 8 << ((control1 >> 5) & 1);						//detect 16 pixel high sprites
	if(control2 & 0x8)
		{
		p = pipeline[0].attr;
		p |= (pipeline[0].p[0] >> n) & 1;
		p |= ((pipeline[0].p[1] >> n) & 1) << 1;
		}
	if(control2 & 0x10)
		{
		for(i=0;i<sprinrange;i++)
			{
			if(sprtemp[i].x == 0)							//is it time to be drawn?
				sprtemp[i].tmpx = 8;							//draw this sprite immediately
			sprtemp[i].x--;									//decrement x value
			if(sprtemp[i].tmpx > 0)							//is this sprite being drawn?
				{
				sprtemp[i].tmpx--;
				n = sprtemp[i].tmpx;
				if(sprtemp[i].attr & 0x40)					//horizontal flip
					n = 7 - sprtemp[i].tmpx;
				if(spr0hit == 0 && sprtemp[i].index == 0)	//sprite 0
					{
					sp = (sprtemp[i].attr & 3) << 2;
					sp |= (sprtemp[i].p[0] >> n) & 1;
					sp |= ((sprtemp[i].p[1] >> n) & 1) << 1;
					if((p & 3) && (sp & 3))
						{
						if(((control2 & 0x1E) != 0x1E) && (pixel < (8+finex)) || pixel == 255)
							{
							//message("sprite0 hit clipped: pixel = %d, ppumask = $%02X\n",pixel,control2);
							}
						else
							{
							//message("sprite0 hit: pixel = %d, ppumask = $%02X\n",pixel,control2);
							status |= 0x40;
							spr0hit = 1;
							}
						}
					}
				if(sprtemp[i].attr & 0x20)					//is sprite behind background?
					{
					if((p & 3) == 0)							//is this transparent pixel?
						{
						sp = (sprtemp[i].attr & 3) << 2;
						sp |= (sprtemp[i].p[0] >> n) & 1;
						sp |= ((sprtemp[i].p[1] >> n) & 1) << 1;
						}
					}
				else
					{
					sp = (sprtemp[i].attr & 3) << 2;
					sp |= (sprtemp[i].p[0] >> n) & 1;
					sp |= ((sprtemp[i].p[1] >> n) & 1) << 1;
					}
				if(sp & 3)
					p = sp | 0x10;
				}
			}
		}
	if(control2 & 1)
		p &= 0x30;
	if((control2 & 0x1E) != 0x1E && (pixel < (8+finex)))
		sp = 0;
	if((control2 & 2) == 0 && (pixel < (8+finex)))
		linepixels[pixel] = pal32[(control2 >> 5) & 7][palram[sp | 0x10]];
	else
		linepixels[pixel] = pal32[(control2 >> 5) & 7][palram[p]];
	}
pixel++;
framecycles++;
if(cycle)
	cycle(pixel);
}

void Cppu::Run(int c)
{
int i;

for(i=0;i<c;i++)
	Step();
}

u8 Cppu::Read(u32 addr)
{
u8 ret = latch;

switch(addr & 7)
	{
	case 2:
//		message("$2002 read: pixel = %d, scanline = %d\n",pixel,scanline);
		ret = status | (latch & 0x1F);
		status &= ~0x80;
		toggle = 0;
		break;
	case 4:
		ret = sprmem[spraddr];
		break;
	case 7:
		ret = latch;
		latch = MemRead(ppuaddr & 0x3FFF);
		if((ppuaddr & 0x3F00) == 0x3F00)
			ret = palram[ppuaddr & 0x1f];
		if(control1 & 4)
			ppuaddr += 0x20;
		else
			ppuaddr++;
		ppuaddr &= 0x3FFF;
		break;
	}
//message("ppu read; reg $%04X, returned $%02X\n",addr&7,ret);
return(ret);
}

void Cppu::Write(u32 addr,u8 data)
{
//message("ppu write; reg $%04X = $%02X\n",addr&7,data);
//latch = data;
switch(addr & 7)
	{
	case 0:
		pputemp &= ~(3 << 10);
		pputemp |= (data & 3) << 10;
		control1 = data;
		//message("control1 changed on pixel %d line %d to $%02X\n",pixel,scanline,control1);
		break;
	case 1:
		control2 = data;
		//message("control2 changed on pixel %d line %d to $%02X\n",pixel,scanline,control2);
		break;
	case 3:
		spraddr = data;
		break;
	case 4:
		sprmem[spraddr++] = data;
		break;
	case 5:
		if(toggle == 0)
			{
			pputemp &= ~0x1F;
			pputemp |= data >> 3;
			finex = data & 7;
			toggle++;
			}
		else
			{
			pputemp &= ~((0x1F << 5) | (7 << 12));
			pputemp |= (data & ~7) << 2;
			pputemp |= (data & 7) << 12;
			toggle = 0;
			}
		break;
	case 6:
		if(toggle == 0)
			{
			pputemp &= 0xFF;
			pputemp |= (data & 0x3F) << 8;
			toggle++;
			}
		else
			{
			pputemp &= ~0xFF;
			pputemp |= data;
			ppuaddr = pputemp;
			toggle = 0;
			}
		break;
	case 7:
		ppuaddr &= 0x3FFF;
		if(ppuaddr <= 0x2000)
			MemWrite(ppuaddr,data);
		else if(ppuaddr < 0x3F00)
			MemWrite(ppuaddr & 0x2FFF,data);
		else if(ppuaddr >= 0x3F00)
			{
			data &= 0x3f;
/*			if((ppuaddr & 3) == 0)
				{
				palram[ppuaddr & 0xF] = data;
				palram[(ppuaddr & 0xF) | 0x10] = data;
				}
			else
				palram[ppuaddr & 0x1f] = data;*/
			if((ppuaddr & 0x0f) == 0)
				{
				int i;

				for(i=0;i<8;i++)
					palram[i * 4] = data;
				}
			else if(ppuaddr & 3)
				palram[ppuaddr & 0x1f] = data;
			}
		if(control1 & 4)
			ppuaddr += 32;
		else
			ppuaddr++;
		break;
	}
}
