#ifndef ppuH
#define ppuH

#include "types.h"

typedef struct
	{
	u8 nt;												//nametable byte
	u32 chraddr;										//precalculated chr address
   u8 attr;												//attrib bits << 2
   u8 p[2];												//pattern bytes
   } ppupixel_t;										//pixel pipeline

typedef struct
	{
   u8 p[2];												//pattern bytes
   u8 attr;												//attrib bits << 2
	u8 x;													//x coord
	u8 tmpx;												//x used when drawing
	u8 index;											//index in sprite memory
	u8 tile;												//tile number
   } sprtemp_t;										//sprite temp entry

#define MIRROR_1L	10
#define MIRROR_1H	11

typedef u8 (*ppureadfunc)(u32);
typedef void (*ppuwritefunc)(u32,u8);

class Ccpu;
class Cppu
	{
   public:
	u16 pal16[8][256];								//16 bit palettes
	u32 pal32[8][256];								//32 bit palettes
	u32 *scr32;											//pointer to screen
	int pitch;											//pitch of screen array
   u8 ntram[0x1000];									//two built-in nametables + two extra
   u8 palram[32];										//palette
	u8 sprmem[256];									//sprite memory
	sprtemp_t sprtemp[64];							//temporary sprite memory for scanline
   u8 *readptr[16];									//16 1k pages
   u8 *writeptr[16];									//16 1k pages
	ppureadfunc readfunc[16];						//read function pointers
	ppuwritefunc writefunc[16];					//write function pointers
   ppupixel_t pipeline[3];							//0=cur pixel, 1,next, 2,nextnext
   int scanline,pixel;								//current scanline/pixel
	u64 frames;											//number of frames executed
	u32 framecycles;
   u8 status;											//status register
   u8 control1,control2;							//control registers
   u32 temp;											//ppu current address
   u32 ppuaddr;										//internal address
   u32 pputemp;										//latch, info from loopy's doc
   u8 finex;											//fine x scroll register
	u8 spraddr;											//sprmem address
	u8 latch;											//latch
	u8 sprinrange;										//internal sprite counter
	int spr0hit;										//has sprite 0 been hit this frame
   int toggle;											//$2005/$2006 toggle
   int evenodd;										//ppu even/odd frame flag
	u32 linepixels[256+8];							//line data rendered to tv
	void (*cycle)(int pixel);						//callback on each cycle
   Ccpu *cpu;
   Cppu();
	void GeneratePalettes();
	void SetNT(int nt,int bank);
	void SetMirroring(int m);
   void Reset();
   void GetSprites();
   void Step();
   void Run(int c);
	u8 Read(u32 addr);
	void Write(u32 addr,u8 data);
   u8 MemRead(u32 addr);
   void MemWrite(u32 addr,u8 data);
   };

#include "cpu.h"

#endif
