#include <windows.h>
#include <commctrl.h>
#include <shlobj.h>
#include <stdio.h>
#include "romsdlg.h"
#include "resource.h"
#include "main.h"
#include "list.h"
#include "mappers.h"

#include "crc32.h"

#define NUMCOLUMNS	3

typedef struct rominfo_s
	{
	char filename[MAX_PATH];
	int mapper;
	int compatable;
	} rominfo_t;

class romlist_c:public list_c
	{
	private:
		HWND listview;
		int sortcol,unread;
		static int CALLBACK filenamecomparefunc(LPARAM lParam1,LPARAM lParam2,LPARAM lParamSort);
		static int CALLBACK mappercomparefunc(LPARAM lParam1,LPARAM lParam2,LPARAM lParamSort);
		static int CALLBACK compatablecomparefunc(LPARAM lParam1,LPARAM lParam2,LPARAM lParamSort);
		static void addromfile(char *filename,void *userdata);
		void refreshrom(int index,rominfo_t *rominfo); //refreshes rominfo for listview item index
	public:
		romlist_c(HWND listviewcontrol,int sortc):list_c(){listview=listviewcontrol;sortcol=sortc;}
		rominfo_t *get(int n){return((rominfo_t*)list_c::get(n));}
		int sortcolumn(int ncol = -1){if(ncol>-1)sortcol=ncol;return(sortcol);}
		void init();
		int message(HWND dlgwnd,UINT imsg,WPARAM wparam,LPARAM lparam); //control message
		void refresh(char *romdir); //refreshes rom filenames (or builds new list)
		void list(); //list the roms
		void sort(int column = -1); //sort the list control by column (if given)
	};

static void FindFiles(void(*ProccessFile)(char*,void*),char *ext,void *userdata) //thanks proxy
{
WIN32_FIND_DATA wfd;
HANDLE hFile;
char strTmp[_MAX_PATH];
char dir[512];

GetCurrentDirectory(512,dir);
memset(&wfd,0,sizeof(WIN32_FIND_DATA));
GetCurrentDirectory(_MAX_PATH,strTmp);
strcat(strTmp,"\\");
strcpy(strTmp,ext);
hFile = FindFirstFile(strTmp,&wfd);
if(hFile == NULL || strcmp(wfd.cFileName,"") == 0)
	return;
do
	{
	if((wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
		{
		char pathname[1024];

		wsprintf(pathname,"%s\\%s",dir,wfd.cFileName);
		ProccessFile(pathname,userdata);
		}
	} while(FindNextFile(hFile,&wfd));
FindClose(hFile);
}

int CALLBACK romlist_c::filenamecomparefunc(LPARAM lParam1,LPARAM lParam2,LPARAM lParamSort)
{
romlist_c *romlist = (romlist_c*)lParamSort;

return(strcmpi(romlist->get(lParam1)->filename,romlist->get(lParam2)->filename));
}

int CALLBACK romlist_c::mappercomparefunc(LPARAM lParam1,LPARAM lParam2,LPARAM lParamSort)
{
romlist_c *romlist = (romlist_c*)lParamSort;
int m1,m2;

m1 = romlist->get(lParam1)->mapper;
m2 = romlist->get(lParam2)->mapper;
if(m1 == m2) //same mapper so use filename to sort by
	return(filenamecomparefunc(lParam1,lParam2,lParamSort));
return(m1 - m2);
}

int CALLBACK romlist_c::compatablecomparefunc(LPARAM lParam1,LPARAM lParam2,LPARAM lParamSort)
{
romlist_c *romlist = (romlist_c*)lParamSort;
int c1,c2;

c1 = romlist->get(lParam1)->compatable;
c2 = romlist->get(lParam2)->compatable;
if(c1 == c2)
	return(filenamecomparefunc(lParam1,lParam2,lParamSort));
return(c1 - c2);
}

void romlist_c::addromfile(char *filename,void *userdata)
{
list_c *romlist = (list_c*)userdata;
rominfo_t *rominfo = new rominfo_t;

strcpy(rominfo->filename,filename);
rominfo->mapper = -1;
rominfo->compatable = -1;
romlist->add(rominfo);
}

static int isgoodmapper(char *fn,int m)
{
return(MapperSupported(m));
}

void romlist_c::refreshrom(int ctrlindex,rominfo_t *rominfo)
{
FILE *fp;
u8 header[16];
LVITEM item,subitem1,subitem2;
char mappernum[32];

if((fp = fopen(rominfo->filename,"rb")) == 0)
	return;
fread(&header,16,1,fp);
fclose(fp);
rominfo->mapper = ((header[6] & 0xf0) >> 4) | (header[7] & 0xf0);
rominfo->compatable = isgoodmapper(rominfo->filename,rominfo->mapper)==0?1:0;
sprintf(mappernum,"%d",rominfo->mapper);
item.mask = LVIF_PARAM;
item.iItem = ctrlindex;
item.iSubItem = 0;
ListView_GetItem(listview,&item);
memset(&subitem1,0,sizeof(LVITEM));
memset(&subitem2,0,sizeof(LVITEM));
subitem1.mask = LVIF_TEXT;
subitem1.iItem = ctrlindex;
subitem1.iSubItem = 1;
subitem1.pszText = mappernum;
subitem1.cchTextMax = strlen(subitem1.pszText) + 1;
subitem2.mask = LVIF_TEXT;
subitem2.iItem = ctrlindex;
subitem2.iSubItem = 2;
subitem2.pszText = rominfo->compatable==0?"no":"yes";
subitem2.cchTextMax = strlen(subitem2.pszText) + 1;
ListView_SetItem(listview,&subitem1);
ListView_SetItem(listview,&subitem2);
}

void romlist_c::init()
{
int iCol,numcols = NUMCOLUMNS;
LVCOLUMN pcols[NUMCOLUMNS];
char *columnnames[NUMCOLUMNS] = {"filename","mapper","compatable?"};

memset(pcols,0,sizeof(LVCOLUMN) * NUMCOLUMNS);
pcols[0].mask =
pcols[1].mask =
pcols[2].mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
pcols[0].cx = 200;
pcols[1].cx = 50;
pcols[2].cx = 73;
for(iCol=0;iCol<numcols;iCol++)
	{
	pcols[iCol].fmt = LVCFMT_LEFT;
	pcols[iCol].pszText = columnnames[iCol];
	pcols[iCol].cchTextMax = strlen(pcols[iCol].pszText) + 1;
	pcols[iCol].iOrder = iCol;
	ListView_InsertColumn(listview,iCol,&pcols[iCol]);
	}
}

int romlist_c::message(HWND dlgwnd,UINT imsg,WPARAM wparam,LPARAM lparam)
{
NMHDR *nmhdr;
NMLISTVIEW *nmlv;
NMLVCUSTOMDRAW *nmlvcd;
rominfo_t *rominfo;
LVITEM item;
FILE *fp;
u8 hdr[16];
char str[128],*p;

switch(imsg)
	{
	case WM_NOTIFY:
		nmhdr = (NMHDR*)lparam;
		switch(nmhdr->code)
			{
			case LVN_ITEMCHANGED:
				nmlv = (NMLISTVIEW*)lparam;
				item.mask = LVIF_PARAM;
				item.iItem = nmlv->iItem;
				item.iSubItem = 0;
				ListView_GetItem(listview,&item);
				p = ((rominfo_t*)get(item.lParam))->filename;
				if((fp = fopen(p,"rb")) != 0)
					{
					long nn;
					u8 *shithole;

					fread(&hdr,16,1,fp);
					nn = (hdr[4] * 0x4000) + (hdr[5] * 0x2000) + 16;
					shithole = new u8[nn+1];
					fseek(fp,SEEK_SET,0);
					fread(shithole,nn,1,fp);
					chksum_crc32gentab();
					nn = chksum_crc32(shithole,nn);
					delete[] shithole;
					sprintf(str,"$%08X",nn);
					SetDlgItemText(dlgwnd,IDC_CRCSTATIC,str);
					sprintf(str,"PRG: %dk (%d 16k banks)",(hdr[4] * 0x4000) / 1024,hdr[4]);
					SetDlgItemText(dlgwnd,IDC_PRGSIZESTATIC,str);
					sprintf(str,"CHR: %dk (%d 8k banks)",(hdr[5] * 0x2000) / 1024,hdr[5]);
					SetDlgItemText(dlgwnd,IDC_CHRSIZESTATIC,str);
					sprintf(str,"Mapper %d",((hdr[6] & 0xF0) >> 4) | (hdr[7] & 0xF0));
					SetDlgItemText(dlgwnd,IDC_MAPPERSTATIC,str);
					fclose(fp);
					}
				else
					::message("error loading rom info for '%s'\n",p);
				return(0);
			case NM_DBLCLK:
				nmlv = (NMLISTVIEW*)lparam;
				item.mask = LVIF_PARAM;
				item.iItem = nmlv->iItem;
				item.iSubItem = 0;
				ListView_GetItem(listview,&item);
				if(nes.LoadROM(((rominfo_t*)get(item.lParam))->filename) == 0)
					{
					EndDialog(dlgwnd,0);
					nes.Reset();
					}
				else
					::message("error loading rom '%s'\n",((rominfo_t*)get(item.lParam))->filename);
				return(0);
			case NM_CUSTOMDRAW:
				nmlvcd = (NMLVCUSTOMDRAW*)lparam;
				switch(nmlvcd->nmcd.dwDrawStage)
					{
					case CDDS_PREPAINT:
						SetWindowLong(dlgwnd,DWL_MSGRESULT,CDRF_NOTIFYITEMDRAW);
						return(CDRF_NOTIFYITEMDRAW);
					case CDDS_ITEMPREPAINT:
						rominfo = (rominfo_t*)get(nmlvcd->nmcd.lItemlParam);
						if(rominfo->mapper == -1 || rominfo->compatable == -1)
							{
							refreshrom(nmlvcd->nmcd.dwItemSpec,rominfo);
							unread--;
							}
						SetWindowLong(dlgwnd,DWL_MSGRESULT,CDRF_DODEFAULT);
						return(CDRF_DODEFAULT);
					}
				return(CDRF_DODEFAULT);
			case LVN_COLUMNCLICK:
				nmlv = (NMLISTVIEW*)lparam;
				if(unread && nmlv->iSubItem > 0)
					{
					int i;

					for(i=0;i<num();i++)
						{
						rominfo = (rominfo_t*)get(i);
						if(rominfo->mapper == -1 || rominfo->compatable == -1)
							{
							LVFINDINFO lvfi;

							lvfi.flags = LVFI_PARAM;
							lvfi.lParam = i;
							refreshrom(ListView_FindItem(listview,-1,&lvfi),rominfo);
							unread--;
							}
						}
					}
				sort(nmlv->iSubItem);
				return(0);
			}
		break;
	}
return(0);
}

void romlist_c::refresh(char *romdir)
{
char cd[MAX_PATH];

GetCurrentDirectory(MAX_PATH,cd);
SetCurrentDirectory(romdir);
clear();
FindFiles(addromfile,"*.nes",(void*)this);
unread = num();
SetCurrentDirectory(cd);
list();
sort();
}

void romlist_c::list()
{
int i;
LVITEM item,subitem1,subitem2;
char mappernum[32];

ListView_DeleteAllItems(listview);
for(i=0;i<num();i++)
	{
	rominfo_t *ri = (rominfo_t*)get(i);

	sprintf(mappernum,"%d",ri->mapper);
	memset(&item,0,sizeof(LVITEM));
	memset(&subitem1,0,sizeof(LVITEM));
	memset(&subitem2,0,sizeof(LVITEM));
	item.mask = LVIF_TEXT | LVIF_PARAM;
	item.iItem = i;
	item.pszText = strrchr(ri->filename,'\\') + 1;
	item.cchTextMax = strlen(item.pszText) + 1;
	item.lParam = i;
	subitem1.mask = LVIF_TEXT;
	subitem1.iItem = i;
	subitem1.iSubItem = 1;
	subitem1.pszText = mappernum;
	subitem1.cchTextMax = strlen(subitem1.pszText) + 1;
	subitem2.mask = LVIF_TEXT;
	subitem2.iItem = i;
	subitem2.iSubItem = 2;
	subitem2.pszText = ri->compatable==0?"no":"yes";
	subitem2.cchTextMax = strlen(subitem2.pszText) + 1;
	ListView_InsertItem(listview,&item);
	ListView_SetItem(listview,&subitem1);
	ListView_SetItem(listview,&subitem2);
	}
}

void romlist_c::sort(int column)
{
if(column == -1)
	column = sortcol;
sortcol = column;
switch(sortcol)
	{
	case 0:ListView_SortItems(listview,filenamecomparefunc,this);break;
	case 1:ListView_SortItems(listview,mappercomparefunc,this);break;
	case 2:ListView_SortItems(listview,compatablecomparefunc,this);break;
	}
}

static int CALLBACK browsecallback(HWND hwnd,UINT uMsg,LPARAM lParam,LPARAM lpData)
{
if(uMsg == BFFM_INITIALIZED)
	SendMessage(hwnd,BFFM_SETSELECTION,TRUE,(LPARAM)lpData);
else if(uMsg == BFFM_SELCHANGED)
	{}
return(0);
}

static int dirdialog(HWND wnd,char *dest,char *title)
{
BROWSEINFO bi;
char dir[MAX_PATH];
char defaultpath[MAX_PATH];
LPITEMIDLIST pidl;

strcpy(defaultpath,dest);
if(strcmp(defaultpath,"") == 0)
	strcpy(defaultpath,curdir);
bi.hwndOwner = wnd;
bi.pidlRoot = 0;
bi.pszDisplayName = dir;
bi.lpszTitle = title;
bi.ulFlags = BIF_EDITBOX;
bi.lpfn = browsecallback;
bi.lParam = (LPARAM)(char*)defaultpath;
bi.iImage = 0;
if((pidl = SHBrowseForFolder(&bi)) == 0)
	return(1);
else
	{
	LPMALLOC pMalloc;

	SHGetPathFromIDList(pidl,dest);
	SHGetMalloc(&pMalloc);
	pMalloc->Release();
	}
return(0);
}

static int CALLBACK romsdlgproc(HWND dlgwnd,UINT imsg,WPARAM wparam,LPARAM lparam)
{
static romlist_c *romlist = 0;
char romdir[512];
char str[128];

switch(imsg)
	{
	case WM_INITDIALOG:
		strncpy(romdir,curdir,512);
		if(romlist)
			delete romlist;
		romlist = new romlist_c(GetDlgItem(dlgwnd,IDC_ROMLIST),0);
		romlist->init();
		romlist->refresh(romdir);
		sprintf(str,"Click an image to get its info.");
		SetDlgItemText(dlgwnd,IDC_CRCSTATIC,(LPCSTR)str);
		sprintf(str,"Double click an image to play.");
		SetDlgItemText(dlgwnd,IDC_PRGSIZESTATIC,(LPCSTR)str);
		str[0] = 0;
		SetDlgItemText(dlgwnd,IDC_CHRSIZESTATIC,(LPCSTR)str);
		SetDlgItemText(dlgwnd,IDC_MAPPERSTATIC,(LPCSTR)str);
		SetDlgItemText(dlgwnd,IDC_DIRECTORYEDIT,(LPCSTR)romdir);
		return(1);
	case WM_NOTIFY:
		if(wparam == IDC_ROMLIST)
			return(romlist->message(dlgwnd,imsg,wparam,lparam));
		break;
	case WM_COMMAND:
		switch(wparam)
			{
			case IDC_REFRESH:
				romlist->refresh(romdir);
				return(1);
			case IDC_BROWSEBTN:
				strncpy(romdir,curdir,512);
				if(dirdialog(dlgwnd,romdir,"Browse for image directory") == 0)
					{
					SetDlgItemText(dlgwnd,IDC_DIRECTORYEDIT,romdir);
					romlist->refresh(romdir);
					strncpy(curdir,romdir,512);
					}
				return(1);
			case IDC_GOBTN:
				GetDlgItemText(dlgwnd,IDC_DIRECTORYEDIT,romdir,512);
				romlist->refresh(romdir);
				strncpy(curdir,romdir,512);
				return(1);
			case IDOK:
			case IDCANCEL:
				if(romlist)
					delete romlist;
				romlist = 0;
				EndDialog(dlgwnd,0);
				return(1);
			}
		break;
	}
return(0);
}

void romsdlg(HINSTANCE hinstance,HWND hwnd)
{
DialogBox(hinstance,MAKEINTRESOURCE(IDD_ROMDLG),hwnd,romsdlgproc);
}
